#!/usr/bin/env python

__version__ = "0.1.15"
__author__ = "Deren Eaton"

from .Toytree import Toytree as tree
from .utils import RandomTree as rtree
from .Multitree import MultiTree as mtree
from .TreeStyle import COLORS as colors

#from .multitree import MultiTree as multitree
#from .tree import Toytree as tree
#from .rtree import RandomTree as rtree
#from .style import COLORS as colors
# from .tree import 